from typing import Union, List, Match, Optional, Iterable, Deque

import re
from collections import deque

from swimport.swim import Swim
from swimport.pools import Pool, syspools
from swimport.model import NameTrigger, Function, object_swimporting, Compileable, TriggerBehaviourSwimRule, SwimRule
from swimport.__util__ import *

MACRO_PATTERN = re.compile(r'\$({(?P<k_b>[^}]+)}|(?P<k>[a-zA-Z_0-9]+)|(?P<other>({}|.)))')
macro_dict = {'retvar': 'val', }


# todo python:cdefaultargs?

def apply_pycode_macros(code, env_dict=None, **kwargs):
    """apply special macros for python code"""
    d = env_dict
    if not d:
        d = kwargs
    elif kwargs:
        d = {**kwargs, **env_dict}

    def macro(match: Match[str]):
        match_string = match['k']
        if match_string:
            return macro_dict[match_string]
        match_string = match['k_b']
        if match_string:
            return str(eval(match_string, d))
        match_string = match['other']
        if match_string:
            if match_string == '$':
                return '$'
            raise ValueError('cannot match template code: $' + match_string)
        raise Exception('unrecognized match: ' + match.group())

    return MACRO_PATTERN.sub(macro, code)


@Function.Param.set_default_trigger
class ParameterNameTrigger(NameTrigger, Function.Param.Trigger):
    pass


class ParameterTypeTrigger(Function.Param.Trigger):
    """
    Trigger for specific parameters types.
    """

    def __init__(self, pattern: Compileable):
        self.pattern = re.compile(pattern)

    def is_valid(self, rule: 'SwimRule', obj: Function.Param):
        return super().is_valid(rule, obj) \
               and self.pattern.fullmatch(obj.normalized_type)

    const_ptr: 'ParameterTypeTrigger'
    not_const_ptr: 'ParameterTypeTrigger'


class ParameterBehaviour(Function.Param.Behaviour):
    """
    Trigger for a function parameter
    NOTE: unlike regular behaviours, this behaviour returns a ParameterUsageInfo.
    ParameterBehaviour object should not be used as a regular behaviour, since ParameterUsageInfo is not a Swimporting.
    """

    class ParameterUsageInfo:
        """An object representing how a parameter should be used"""

        def __init__(self):
            self.params_used: List[Function.Param] = []
            self.pre_decl_scopes: List = []
            self.post_decl_scopes: List = []
            self.pools = []

        def __bool__(self):
            return bool(self.params_used or self.pre_decl_scopes or self.post_decl_scopes or self.pools)

    def wrap(self, rule, param) -> ParameterUsageInfo:
        ret = self.ParameterUsageInfo()
        if param:
            ret.params_used.append(param)
        return ret


class ParameterRule(TriggerBehaviourSwimRule):
    def do(self, obj: Function.parameters) -> Optional[ParameterBehaviour.ParameterUsageInfo]:
        return super().do(obj)


ParameterBehaviour.rule_cls = ParameterRule


@Function.Param.set_default_behaviour
class ApplyBehaviour(ParameterBehaviour):
    """Applies keywords from another typemap to the parameter or parameters"""

    def __init__(self, *keywords: str, pools: Iterable[Pool] = ()):
        """
        :param keywords: rthe keywords to apply to the parameters, the number of keywords is the number of parameters consumed
        """
        self.keywords = keywords
        self.pools = pools

    def wrap(self, rule, param: Function.Param):
        ret = super().wrap(rule, None)
        ret.pools.extend(self.pools)
        if self.keywords:
            explanation = '// rule: ' + str(rule)
            copy_parts = []
            sign_parts = []
            for kw in self.keywords:
                if param is None:
                    # if we ran out of parameters unexpectedly, fail by returning an empty parameter usage info
                    return self.ParameterUsageInfo()
                ret.params_used.append(param)
                copy_parts.append(f'{param.type} {kw}')
                sign_parts.append(param.sign)
                param = param.next_or_none
            if len(copy_parts) == 1:
                copy_parts = copy_parts[0]
            else:
                copy_parts = '(' + ', '.join(copy_parts) + ')'

            if len(sign_parts) == 1:
                sign_parts = sign_parts[0]
            else:
                sign_parts = '(' + ', '.join(sign_parts) + ')'
            ret.pre_decl_scopes.append(
                Swim.add_raw(f'%apply {copy_parts} {{{sign_parts}}};' + explanation)
            )
            ret.post_decl_scopes.append(
                Swim.add_raw(f'%clear {sign_parts};' + explanation)
            )
        return ret


# region Triggers
# region NameTriggers
ParameterNameTrigger.ref_io = ParameterNameTrigger('IO_.*')
# endregion
# region TypeTriggers
ParameterTypeTrigger.ptr = ParameterTypeTrigger(r'.*[&*]')
ParameterTypeTrigger.const_ptr = ParameterTypeTrigger(r'.* const [&*]')
ParameterTypeTrigger.not_const_ptr = ParameterTypeTrigger.ptr & ~ParameterTypeTrigger.const_ptr
# endregion
# endregion
# region behaviours
ParameterBehaviour.default = ParameterBehaviour()
ParameterBehaviour.output = ApplyBehaviour('OUTPUT')
ParameterBehaviour.input = ApplyBehaviour('INPUT')
ParameterBehaviour.inout = ApplyBehaviour('INOUT')
# endregion
# region rules

ParameterRule.inout = ParameterRule(ParameterNameTrigger.ref_io,
                                    ParameterBehaviour.inout, 'inout')
ParameterRule.output = ParameterRule(ParameterTypeTrigger.not_const_ptr,
                                     ParameterBehaviour.output, 'out')
ParameterRule.input = ParameterRule(ParameterTypeTrigger.const_ptr,
                                    ParameterBehaviour.input, 'input')
ParameterRule.default = ParameterBehaviour.default


# endregion

@Function.set_default_behaviour
class FunctionBehaviour(Function.Behaviour):
    default_parameters_rules: Deque[ParameterRule] = deque((
        # regular
        ParameterRule.inout, ParameterRule.output,
        ParameterRule.input,
        # default
        ParameterRule.default
    ))

    def __init__(self, parameter_rules: List[ParameterRule] = ..., dll_detect=(),
                 prepend_python: str = None, append_python: str = None, exception_check=None, contract=None,
                 free_ret: Union[str, bool] = False, autodoc: Union[str, int, None] = 2,
                 docstring: str = None, body: str = '', body_scope: str = ...):
        """
        :param parameter_rules: rules for handling parameters
        :param dll_detect: dll detection dictionary
        :param prepend_python: python code to run before function is called
        :param append_python: python code to run after function is called
        :param exception_check: cpp code to check for and handle exceptions (see %exception in SWIG).
         use ... to check for all errors.
        :param contract: cpp contract for function (see %contract in SWIG)
        :param free_ret: whether to insert a clause to free the output of the function. set to True for delete,
        or specify a releasing function.
        :param autodoc: autodoc value for function (see %autodoc in SWIG)
        :param docstring: docstring for the function
        """
        if parameter_rules is ...:
            parameter_rules = self.default_parameters_rules
        self.parameter_rules: List[ParameterRule] = list(parameter_rules)
        self.dll_detect = dict(dll_detect)
        self.prepend = prepend_python
        self.append = append_python
        self.exception = exception_check
        self.contract = contract
        self.free_ret = free_ret
        self.autodoc = autodoc
        self.docstring = docstring
        self.body = body
        self.body_scope = body_scope

        if self.body:
            assert not self.dll_detect, 'function body must not be used with dll detect'

    @object_swimporting
    def wrap(self, rule, obj: Function, swim):
        post_decl_scopes = []
        pre_decl_scopes = []
        debug = {'parameter rules': [], 'dll name': []}

        swim.add_comment(f'{obj.name} ({rule})')

        body = self.body
        decl_slice = slice(0, None)

        # check for dll
        for dll_prefix, dll_name in self.dll_detect.items():
            if obj.return_type.startswith(dll_prefix):
                if dll_name.endswith('.dll'):
                    dll_name = dll_name[:-len('.dll')]
                clean_rtn_type = obj.return_type[len(dll_prefix):]
                decl_slice = slice(decl_slice.start + len(dll_prefix), decl_slice.stop, decl_slice.step)
                debug['dll name'].append(dll_name)
                swim.add_insert(
                    f'SWIM_DECL_IMPORTED_FUNC({clean_rtn_type},{obj.name}'
                    f', {", ".join(obj.prm_types)})')
                swim.add_init(f"SWIM_IMPORT_FROMDLL({obj.name}, {dll_name})")
                # note: if dll_detect exists, body must be initially empty
                body = '{\n\t' \
                       + ('' if clean_rtn_type.rstrip().endswith('void') else 'return ') \
                       + 'SWIM_FUNC_HANDLE_' + obj.name + '(' + ', '.join(obj.prm_names) + ');\n}'
                break

        # run through and assign rules to all the parameters
        next_param_ind = 0
        while len(obj.parameters) > next_param_ind:
            param = obj.parameters[next_param_ind]
            for param_rule in self.parameter_rules:
                usage = param_rule.do(param)
                if not usage:
                    continue

                for pool in usage.pools:
                    swim(pool)
                pre_decl_scopes.extend(usage.pre_decl_scopes)
                post_decl_scopes.extend(usage.post_decl_scopes)
                if not usage.params_used:
                    continue

                next_param_ind += len(usage.params_used)
                break
            else:
                raise Exception('parameter without rule: ' + param.body + ' at method ' + obj.name)

        # add all the special features:
        # region special features
        if self.append:
            append = apply_pycode_macros(self.append, method=obj)
            pre_decl_scopes.append(
                Swim.add_python_append(obj, append)
            )
            post_decl_scopes.append(
                Swim.add_python_append(obj, None)
            )

        if self.prepend:
            prepend = apply_pycode_macros(self.prepend, method=obj)
            pre_decl_scopes.append(
                Swim.add_python_prepend(obj, prepend)
            )
            post_decl_scopes.append(
                Swim.add_python_prepend(obj, None)
            )

        elif self.exception:
            if self.exception is ...:
                swim(syspools.std_exception)
                pre_decl_scopes.append(
                    Swim.add_raw('SWIM_STD_EXCEPTION(' + obj.pattern + ');')
                )
            else:
                exception = clean_source(self.exception)
                pre_decl_scopes.append(
                    Swim.add_exception_check(exception)
                )

            post_decl_scopes.append(Swim.add_raw('%exception;'))

        if self.contract:
            contract = clean_source(self.contract)
            pre_decl_scopes.append(
                Swim.add_contract(obj, contract)
            )
            post_decl_scopes.append(
                Swim.add_contract(obj, None)
            )

        if self.free_ret:
            free_out = self.free_ret
            if free_out is True:
                pre_decl_scopes.append(
                    Swim.add_raw('%newobject ' + obj.pattern + ';')
                )
            else:
                pre_decl_scopes.append(
                    Swim.add_raw('%typemap(ret) ' + obj.return_type + ' ' + obj.name + '{ ' + free_out + '($1); };')
                )
                post_decl_scopes.append(
                    Swim.add_raw('%typemap(ret) ' + obj.return_type + ' ' + obj.name + ';')
                )

        if self.autodoc is not None:
            pre_decl_scopes.append(
                Swim.add_raw(f'%feature("autodoc", "{self.autodoc}") ' + obj.pattern + ';')
            )
            post_decl_scopes.append(
                Swim.add_raw('%feature("autodoc") ' + obj.pattern + ';')
            )

        if self.docstring is not None:
            pre_decl_scopes.append(
                Swim.add_feature(self.docstring, obj.pattern, 'feature("docstring")')
            )
            post_decl_scopes.append(
                Swim.add_feature(self.docstring, None, 'feature("docstring")')
            )
        # endregion

        for scope in pre_decl_scopes:
            scope(swim)

        # actual deceleration goes here
        decl = obj.body[decl_slice]
        decl_scope = self.body_scope
        if body:
            if decl.endswith(';'):
                decl = decl[:-1]
            decl = decl + '\n' + body
        if decl_scope is ...:
            decl_scope = Swim.add_inline if body else Swim.add_raw
        decl_scope(swim, decl)

        for scope in reversed(post_decl_scopes):
            scope(swim)

        swim.add_comment(((n + ': ' + str(v)) for n, v in debug.items()),
                         verbosity_level=Verbosity.debug)

        swim.add_nl()


class FunctionNameTrigger(NameTrigger, Function.Trigger):
    pass


@Function.set_default_trigger
def default_trigger(arg=None):
    if arg is None:
        return Function.Trigger.inherit_cls()
    return FunctionNameTrigger(arg)



from itertools import chain, product
import os
from typing import Dict

from swimport.pools.pools import pools, syspools, IdiomaticPool


@syspools.add(IdiomaticPool)
def np_main(numpy_dir, *, swim):
    swim.add_begin('#define SWIG_FILE_WITH_INIT')
    path = os.path.join(numpy_dir, "numpy.i")
    swim.add_raw(f'%include "{path}"')
    swim.add_init('import_array();')


@syspools.add(IdiomaticPool)
def np_char(*, swim):
    # the NPY_CHAR value is officially deprecated, but currently, numpy.i only makes use of the
    # PyArray_SimpleNew/FromData functions, instead of the PyArray_New function, making it impossible to use flexible
    # types (string, unicode)
    swim.add_raw('%numpy_typemaps(char, NPY_CHAR, size_t)')
    swim.add_raw('%numpy_typemaps(char, NPY_CHAR, int)')


@syspools.add(IdiomaticPool)
def np_applies(dim, additional=(), const_input=True, ref_out=True, typedefs = (), *, swim):
    typedefs = dict(typedefs)
    def dim_params(pref):
        return (pref + ' DIM' + str(i) for i in range(1, dim + 1))

    for data_type in chain(
            (u + t for u, t in product(('unsigned ', ''), ('int', 'long', 'long long', 'short',))),
            ('float', 'double', 'signed char', 'unsigned char'),
            additional):
        alias = None
        if data_type in typedefs:
            alias = typedefs[data_type]
        for dim_type in ('int', 'size_t'):
            if const_input:
                dp = ", ".join(dim_params(dim_type))
                if alias:
                    swim.add_raw(
                        f'%apply ({alias} * IN_ARRAY{dim}, {dp}) '
                        f'{{ ({data_type} * IN_ARRAY{dim}, {dp}) }};'
                    )
                swim.add_raw(
                    f'%apply ({data_type} * IN_ARRAY{dim}, {dp}) '
                    f'{{ ({data_type} const * IN_ARRAY{dim}, {dp}) }};'
                )
                if dim == 1:
                    if alias:
                        swim.add_raw(
                            f'%apply ({alias} * INPLACE_ARRAY_FLAT, {dim_type} DIM_FLAT) '
                            f'{{ ({data_type} * INPLACE_ARRAY_FLAT, {dim_type} DIM_FLAT) }};'
                        )
                    swim.add_raw(
                        f'%apply ({data_type} * INPLACE_ARRAY_FLAT, {dim_type} DIM_FLAT) '
                        f'{{ ({data_type} const * INPLACE_ARRAY_FLAT, {dim_type} DIM_FLAT) }};'
                    )

            if ref_out:
                for out_id in (f'ARGOUTVIEW_ARRAY{dim}', f'ARGOUTVIEWM_ARRAY{dim}'):
                    dp = list(dim_params(''))
                    applies = []
                    main_qdp = None
                    for quals in product('*&', repeat=len(dp)):
                        qdp = ', '.join(dim_type + q + d for (q, d) in zip(quals, dp))
                        if not main_qdp:
                            main_qdp = qdp
                        applies.extend((
                            f'({data_type}*& {out_id}, {qdp})',
                            f'({data_type}** {out_id}, {qdp})',
                        ))
                    applies = ', '.join(applies)
                    if alias:
                        swim.add_raw(
                            f'%apply ({alias}** {out_id}, {main_qdp}) {{({data_type}** {out_id}, {main_qdp})}};'
                        )
                    swim.add_raw(
                        f'%apply ({data_type}** {out_id}, {main_qdp}) {{'
                        + applies +
                        '};'
                    )


@pools.add(IdiomaticPool)
def numpy_arrays(numpy_dir='', additionals=(), typedefs = (), allow_char_arrays=False,
                 const_input=True, ref_out=True, max_dim=1, *,
                 swim):
    """
    adds the inserts for np arrays to used (if you also use the stdint pool, remember to also use the npstdint pool)
    :param numpy_dir: the search path for the numpy.i file
    :param additionals: additional dtypes to support.
    :param allow_char_arrays: whether to include typemaps for char arrays.
    :param const_input: whether to allow const pointers  as arrays (for input only)
    :param ref_out: whether to allow ref (&) pointers as arrays (for output only).
    :param max_dim: the maximum number of numpy dimensions to map for.
    """
    typedefs = dict(typedefs) if typedefs else ()
    if max_dim < 1 or max_dim > 4:
        raise ValueError('maxdim must be between 1 and 4 (inclusive)')

    np_main(numpy_dir, swim)
    additionals = list(additionals)
    if allow_char_arrays:
        np_char(swim=swim)
        additionals.append('char')
    if typedefs:
        additionals.extend(typedefs.keys())

    additionals = tuple(additionals)

    for dim in range(1, max_dim + 1):
        swim(np_applies(dim=dim, const_input=const_input, ref_out=ref_out, additional=additionals, typedefs=tuple(typedefs.items())))

    from swimport.functionswim import ParameterNameTrigger, ParameterTypeTrigger, ApplyBehaviour, FunctionBehaviour

    param_rules = (
        (ParameterNameTrigger('A_.*') & ParameterTypeTrigger.const_ptr) >> ApplyBehaviour('IN_ARRAY1', 'DIM1'),
        (ParameterNameTrigger('A_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour('ARGOUTVIEW_ARRAY1',
                                                                                              'DIM1'),
        (ParameterNameTrigger('AF_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour('ARGOUTVIEWM_ARRAY1',
                                                                                               'DIM1'),
        ParameterNameTrigger('AIO_.*') >> ApplyBehaviour('INPLACE_ARRAY_FLAT', 'DIM_FLAT'),
    )
    # the rules are in reverse order since extendleft reverses the iterable

    # note: in theory, default_parameters_rules should only be changed by the user as a convenience.
    # But we don't expect many collisions on the prefix.
    FunctionBehaviour.default_parameters_rules.extendleft(param_rules)

    for i in range(1, max_dim + 1):
        dim_names = [f'DIM{j}' for j in range(1, i + 1)]
        param_rules = (
            (ParameterNameTrigger(f'A{i}_.*') & ParameterTypeTrigger.const_ptr) >> ApplyBehaviour(
                f'IN_ARRAY{i}', *dim_names),
            (ParameterNameTrigger(f'A{i}_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour(
                f'ARGOUTVIEW_ARRAY{i}', *dim_names),
            (ParameterNameTrigger(f'AF{i}_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour(
                f'ARGOUTVIEWM_ARRAY{i}', *dim_names),
        )
        FunctionBehaviour.default_parameters_rules.extendleft(param_rules)


# todo keep? test? do something?
@pools.add(IdiomaticPool)
def numpy_stdint(swim):
    """
    numpy typemaps for stdint types
    """
    swim.add_raw('''
    %numpy_typemaps(int8_t, NPY_BYTE, int)
    %numpy_typemaps(uint8_t, NPY_UBYTE, int)
    %numpy_typemaps(int16_t, NPY_SHORT, int)
    %numpy_typemaps(uint16_t, NPY_USHORT, int)
    %numpy_typemaps(int32_t, NPY_INT, int)
    %numpy_typemaps(uint32_t, NPY_UINT, int)
    %numpy_typemaps(int64_t, NPY_LONG, int)
    %numpy_typemaps(uint64_t, NPY_ULONG, int)
    %numpy_typemaps(int64_t, NPY_LONGLONG , int)
    %numpy_typemaps(uint64_t, NPY_ULONGLONG, int)
    ''')
    swim.add_nl()
    swim.add_raw('''
    %numpy_typemaps(int8_t, NPY_BYTE, size_t)
    %numpy_typemaps(uint8_t, NPY_UBYTE, size_t)
    %numpy_typemaps(int16_t, NPY_SHORT, size_t)
    %numpy_typemaps(uint16_t, NPY_USHORT, size_t)
    %numpy_typemaps(int32_t, NPY_INT, size_t)
    %numpy_typemaps(uint32_t, NPY_UINT, size_t)
    %numpy_typemaps(int64_t, NPY_LONG, size_t)
    %numpy_typemaps(uint64_t, NPY_ULONG, size_t)
    %numpy_typemaps(int64_t, NPY_LONGLONG , size_t)
    %numpy_typemaps(uint64_t, NPY_ULONGLONG, size_t)
    ''')
    swim.add_nl()


@pools.add(IdiomaticPool)
def buffer(output=bytearray, *, swim):
    """
    adds the necessary typemaps to use bytes objects
    :param output: the type of the python object to output when using bytes
    """
    swim.add_raw("""
    %typemap(in, numinputs=1) (unsigned char const* BUFFER_INPUT, size_t DIM1) (Py_buffer buffer)
    {
        buffer.buf = nullptr;
        auto _buffer = &buffer;
        if (PyObject_GetBuffer($input, _buffer, PyBUF_CONTIG_RO) < 0) SWIG_fail;
        if (_buffer->itemsize != sizeof(unsigned char)){
            PyErr_SetString(PyExc_BufferError, "buffer is of the wrong type in argument: $1_name");
            SWIG_fail;
        }
        $1 = (unsigned char*) _buffer->buf;
        $2 = (size_t) _buffer->len;
    }
    %typemap(freearg, match="in") (unsigned char const* BUFFER_INPUT, size_t DIM1)
    {
        if ((buffer$argnum).buf)
            PyBuffer_Release(&buffer$argnum);
    }
    %typemap(typecheck) (unsigned char const* BUFFER_INPUT, size_t DIM1){
        $1 = PyObject_CheckBuffer($input);
    }
    """)
    if output in (bytearray, bytes):
        out_func = {bytearray: 'PyByteArray_FromStringAndSize', bytes: 'PyBytes_FromStringAndSize'}[output]
        swim.add_raw(f"""
            %typemap(in, numinputs=0) (unsigned char** BUFFER_OUTPUT, size_t* DIM1) (unsigned char* resbuffer, size_t reslen)
            {{
                $1 = &resbuffer;
                $2 = &reslen;
            }}
            %typemap(argout) (unsigned char** BUFFER_OUTPUT, size_t* DIM1)
            {{
                $result = SWIG_Python_AppendOutput($result,{out_func}(*(const char**)$1, *$2));
            }}
            %apply (unsigned char** BUFFER_OUTPUT, size_t* DIM1)
            {{
                (unsigned char*& BUFFER_OUTPUT, size_t& DIM1),
                (unsigned char** BUFFER_OUTPUT, size_t& DIM1),
                (unsigned char*& BUFFER_OUTPUT, size_t* DIM1)
            }};
        
            %typemap(in, numinputs=0) (unsigned char** OUTPUT_FREE, size_t* DIM1) (unsigned char* resbuffer, size_t reslen)
            {{
                $1 = &resbuffer;
                $2 = &reslen;
            }}
            %typemap(argout) (unsigned char** OUTPUT_FREE, size_t* DIM1)
            {{
                $result = SWIG_Python_AppendOutput($result,{out_func}(*(const char**)$1, *$2));
                if ($1 && (*$1))
                    delete[] *$1;
            }}
            %apply (unsigned char** OUTPUT_FREE, size_t* DIM1)
            {{
                (unsigned char*& OUTPUT_FREE, size_t& DIM1),
                (unsigned char** OUTPUT_FREE, size_t& DIM1),
                (unsigned char*& OUTPUT_FREE, size_t* DIM1)
            }};
            """)
    elif output is memoryview:
        # todo test?
        swim.add_raw(f"""
            %typemap(in, numinputs=0) (unsigned char** BUFFER_OUTPUT, size_t* DIM1)(unsigned char* resbuffer, size_t reslen)
            {{
                $1 = &resbuffer;
                $2 = &reslen;
            }}
            %typemap(argout) (unsigned char** BUFFER_OUTPUT, size_t* DIM1)
            {{
                $result = SWIG_Python_AppendOutput($result,PyMemoryView_FromMemory(*(const char**)$1, *$2, PyBUF_WRITE));
            }}
            %apply (unsigned char** BUFFER_OUTPUT, size_t* DIM1)
            {{
                (unsigned char*& BUFFER_OUTPUT, size_t& DIM1),
                (unsigned char** BUFFER_OUTPUT, size_t& DIM1),
                (unsigned char*& BUFFER_OUTPUT, size_t* DIM1)
            }};

            %typemap(in, numinputs=0) (unsigned char** OUTPUT_FREE, size_t* DIM1) (unsigned char* resbuffer, size_t reslen)
            {{
                #error OUTPUT_FREE cannot be used with the memoryview
            }}
            """)
    elif not output:
        pass
    else:
        raise TypeError('could not recognize output')

    from swimport.functionswim import ParameterNameTrigger, ParameterTypeTrigger, ApplyBehaviour, FunctionBehaviour

    param_rules = (
        (ParameterNameTrigger('B_.*') & ParameterTypeTrigger.const_ptr) >> ApplyBehaviour('BUFFER_INPUT', 'DIM1'),
        (ParameterNameTrigger('B_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour('BUFFER_OUTPUT', 'DIM1'),
        (ParameterNameTrigger('BF_.*') & ParameterTypeTrigger.not_const_ptr) >> ApplyBehaviour('OUTPUT_FREE', 'DIM1')
    )
    # the rules are in reverse order since extendleft reverses the iterable

    # note: in theory, default_parameters_rules should only be changed by the user as a convenience.
    # But we don't expect many collisions on the prefix.
    FunctionBehaviour.default_parameters_rules.extendleft(param_rules)



from inspect import signature

from swimport.pools.pools import pools, IdiomaticPool, Pool
from swimport.typeswim import TypeSwimporting, FunctionBody, DefaultCppCheck


class TypeSwimportingPool(Pool):
    """
    A pool that wraps a function returns a TypeSwimporting, or an iterable of TypeSwimportings,
    and automatically applies them to the Swim object.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.swim_param = 'swim' in signature(self.__func__).parameters

    def get(self, *args, **kwargs):
        return self.__func__(*self.args, *args, **self.kwargs, **kwargs)

    def apply(self, swim, args, kwargs):
        if self.swim_param:
            kwargs['swim'] = swim
        ret = self.__func__(*self.args, *args, **kwargs, **self.kwargs)
        return swim(ret)


@pools.add(IdiomaticPool)
def void_ptr(*, swim, **type_args):
    """
    adds the typemaps and functions for void* outputs
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    type_args.setdefault('inout_maps', False)

    voidptr_swimporting = TypeSwimporting('void*',
                                          to_py=
                                          FunctionBody("""
                                            return SWIG_NewPointerObj(input, (swig_type_info*)userdata, 0|0);
                                          """, user_data="userdata = $descriptor(void*);")
                                          ,
                                          to_cpp="""
                                            void* ptr;
                                            int res;
                                            res = SWIG_ConvertPtr(input,SWIG_as_voidptrptr(&ptr), 0, 0);
                                            if (!SWIG_IsOK(res)) {
                                                PyErr_SetString(PyExc_TypeError, "bad type");
                                                ok = false;
                                                return nullptr;
                                            }
                                            return ptr;
                                          """,
                                          to_cpp_check="""
                                            void* ret;
                                            int res;
                                            res = SWIG_ConvertPtr(input,SWIG_as_voidptrptr(&ret), 0, 0);
                                            if (!SWIG_IsOK(res)) {
                                                return false;
                                            }
                                            return true;
                                          """,
                                          **type_args
                                          )
    swim(voidptr_swimporting)
    swim.add_raw(f'%apply void const * {{void const * INPUT}};')
    swim.add_nl()


@pools.add(IdiomaticPool)
def c_string(swim, **type_args):
    """
    adds the typemaps for c_strings
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    swim.add_raw('%clear char*;')
    swim.add_raw('%clear char const *;')

    swim(TypeSwimporting('char const *', 'str',
                         to_py="""
                            if (!input){
                                Py_RETURN_NONE;
                            }
                            return PyUnicode_FromString(input);
                            """,
                         to_cpp=FunctionBody("""
                            return (char*)PyUnicode_AsUTF8(input);
                           """),
                         to_cpp_check_precedence="SWIG_TYPECHECK_UNISTRING",
                         **type_args))
    swim(TypeSwimporting('char *', 'str',
                         to_py="""
                            if (!input){
                                Py_RETURN_NONE;
                            }
                            return PyUnicode_FromString(input);
                            """,
                         **type_args))
    swim(TypeSwimporting('wchar_t const *', 'str',
                         to_py="""
                            if (!input){
                                Py_RETURN_NONE;
                            }
                            return PyUnicode_FromWideChar(input, -1);
                            """,
                         to_cpp=FunctionBody("""
                            Py_ssize_t size = 0;
                            return PyUnicode_AsWideCharString(input, &size);
                          """),
                         to_cpp_post="""
                           PyMem_Free((void*)input);
                           """,
                         to_cpp_check_precedence="SWIG_TYPECHECK_STRING",
                         **type_args))
    swim(TypeSwimporting('wchar_t *', 'str',
                         to_py="""
                            if (!input){
                                Py_RETURN_NONE;
                            }
                            return PyUnicode_FromWideChar(input, -1);
                            """,
                         to_cpp="""
                               return PyUnicode_AsWideCharString(input, nullptr);
                              """,
                         to_cpp_post="""
                               PyMem_Free((void*)input);
                               """,
                         to_cpp_check_precedence="SWIG_TYPECHECK_UNISTRING",
                         **type_args))

    to_py = swim.types['char *'].to_py_func

    swim.add_raw(f"""
        %apply char** OUTPUT {{char** OUTPUT_FREE, char*& OUTPUT_FREE, char const ** OUTPUT_FREE, char const *& OUTPUT_FREE}}
        %typemap(argout) char** OUTPUT_FREE, char*& OUTPUT_FREE, char const ** OUTPUT_FREE, char const *& OUTPUT_FREE%{{
            if ($1){{
                int ok = true;
                void* volatile userdata=nullptr;
                {to_py.user_data}
                PyObject* to_add = {to_py.function_name}(*$1, ok, userdata);
                if (!ok || PyErr_Occurred()) SWIG_fail;
                $result = SWIG_Python_AppendOutput($result, to_add);
                delete[] *$1;
            }}
        %}}
        """)

    to_py = swim.types['wchar_t *'].to_py_func
    swim.add_raw(f"""
        %apply wchar_t** OUTPUT {{wchar_t** OUTPUT_FREE, wchar_t*& OUTPUT_FREE, wchar_t const ** OUTPUT_FREE, wchar_t const *& OUTPUT_FREE}}
        %typemap(argout) wchar_t** OUTPUT_FREE, wchar_t*& OUTPUT_FREE, wchar_t const ** OUTPUT_FREE, wchar_t const *& OUTPUT_FREE%{{
            if ($1){{
                int ok = true;
                void* volatile userdata=nullptr;
                {to_py.user_data}
                PyObject* to_add = {to_py.function_name}(*$1, ok, userdata);
                if (!ok || PyErr_Occurred()) SWIG_fail;
                $result = SWIG_Python_AppendOutput($result, to_add);
                delete[] *$1;
            }}
        %}}
        """)

    swim.add_raw("""
        %typemap(newfree) wchar_t*, wchar_t const *, char*, char const*%{
            delete[] $1;
        %}
        """)

    from swimport.functionswim import FunctionBehaviour, ParameterTypeTrigger, ApplyBehaviour, ParameterNameTrigger

    param_rules = (
        (ParameterNameTrigger('SF_.*') & ParameterTypeTrigger(r'(char|wchar_t)( const)? \* [&*]'))
        >> ApplyBehaviour('OUTPUT_FREE'),
        ParameterTypeTrigger(r'(char|wchar_t) const \*') >> ApplyBehaviour('')
    )

    FunctionBehaviour.default_parameters_rules.extendleft(param_rules)


@pools.add(TypeSwimportingPool)
def std_string(**type_args):
    """
    adds the typemaps for std::strings
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    yield TypeSwimporting('std::string', 'str',
                          to_py="""
                             return PyUnicode_FromString(input.c_str());
                             """,
                          to_cpp="""
                             Py_ssize_t size = 0;
                             auto buffer = PyUnicode_AsUTF8AndSize(input, &size);
                             if (!buffer){
                                SWIM_RAISE_TYPE("expected a string, not ", input);
                             }
                             return {buffer , (size_t)size};
                             """,
                          to_cpp_check_precedence='SWIG_TYPECHECK_STRING',
                          **type_args)

    yield TypeSwimporting('std::wstring', 'str',
                          to_py="""
                            return PyUnicode_FromWideChar(input.c_str(), -1);
                            """,
                          to_cpp="""
                            Py_ssize_t size = 0;
                            auto buffer = PyUnicode_AsWideCharString(input, &size);
                            if (!buffer){
                                ok = false;
                                return {};
                            }
                            std::wstring res(buffer, (size_t)size);
                            PyMem_Free((void*)buffer);
                            return res;
                            """,
                          to_cpp_check_precedence='SWIG_TYPECHECK_UNISTRING',
                          **type_args)


@pools.add(TypeSwimportingPool, name='pyObject')
def py_object(**type_args):
    """
    adds the typemaps to use PyObject* arguments
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    return TypeSwimporting('PyObject*', 'object',
                           to_py="return input;",
                           to_cpp="return input;",
                           to_cpp_check="return true;",
                           to_cpp_check_precedence=...,
                           ref_maps=TypeSwimporting.REF_MAPS_NO_NONE_CHECK,
                           **type_args)


_prim_maps = [
    ('signed char', 'PyLong_FromLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_INT8'),
    ('short', 'PyLong_FromLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_INT16'),
    ('int', 'PyLong_FromLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_INT32'),
    ('long', 'PyLong_FromLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_INT64'),
    ('long long', 'PyLong_FromLongLong', 'PyLong_AsLongLong', 'PyLong_Check', 'SWIG_TYPECHECK_INT128'),

    ('unsigned char', 'PyLong_FromUnsignedLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_UINT8'),
    ('unsigned short', 'PyLong_FromUnsignedLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_UINT16'),
    ('unsigned int', 'PyLong_FromUnsignedLong', 'PyLong_AsLong', 'PyLong_Check', 'SWIG_TYPECHECK_UINT32'),
    ('unsigned long', 'PyLong_FromUnsignedLong', 'PyLong_AsUnsignedLong', 'PyLong_Check', 'SWIG_TYPECHECK_UINT64'),
    ('unsigned long long', 'PyLong_FromUnsignedLongLong', 'PyLong_AsUnsignedLongLong', 'PyLong_Check',
     'SWIG_TYPECHECK_UINT128'),

    ('size_t', 'PyLong_FromUnsignedLong', 'PyLong_AsUnsignedLong', 'PyLong_Check', 'SWIG_TYPECHECK_UINT64'),

    ('float', 'PyFloat_FromDouble', 'PyFloat_AsDouble', None, 'SWIG_TYPECHECK_FLOAT'),
    ('double', 'PyFloat_FromDouble', 'PyFloat_AsDouble', None, 'SWIG_TYPECHECK_DOUBLE'),
    ('long double', 'PyFloat_FromDouble', 'PyFloat_AsDouble', None, 'SWIG_TYPECHECK_DOUBLE'),
]


@pools.add(IdiomaticPool)
def primitive(*, additionals=True, swim, blacklist=(), **type_args):
    """
    includes typemaps to convert regular c++ primitives as swimport types
    :param additionals: whether to apply typemaps for iterables and sequences of sequences of primitive types
    :param blacklist: primitive types not to import
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    if isinstance(blacklist, str):
        blacklist = (blacklist,)

    if additionals:
        def additionals(prim):
            iter_types = TypeSwimporting.default_in_iter_types
            for i in iter_types:
                i = i.replace('?', prim)
                swim(pools.input_iterable(i))

            iter_types = TypeSwimporting.default_out_iter_types
            for i in iter_types:
                i = i.replace('?', prim)
                for j in iter_types:
                    swim(pools.iter(i, cpp_name=j))
    else:
        def additionals(x):
            pass

    def import_(port: TypeSwimporting):
        if port.cpp_name in blacklist:
            return
        swim(port)
        additionals(port.cpp_name)

    for prim, topy, tocpp, check, prec in _prim_maps:
        import_(TypeSwimporting(prim,
                                to_py='return ' + topy + '(input);',
                                to_cpp='return ' + tocpp + '(input);',
                                to_cpp_check=('return ' + check + '(input);' if check else DefaultCppCheck()),
                                to_cpp_check_precedence=prec,
                                **type_args))

    import_(TypeSwimporting('char', 'str',
                            to_py="return PyUnicode_FromStringAndSize(&input, 1);",
                            to_cpp="""
                                 Py_ssize_t size = 0;
                                 auto buffer = PyUnicode_AsUTF8AndSize(input, &size);
                                 if (size != 1){
                                    ok = false;
                                    return {};
                                 }
                                 return buffer[0];
                                 """,
                            to_cpp_check_precedence='SWIG_TYPECHECK_CHAR',
                            **type_args))

    import_(TypeSwimporting('wchar_t', 'str',
                            to_py="return PyUnicode_FromWideChar(&input, 1);",
                            to_cpp="""
                                wchar_t buffer[2];
                                 if (PyUnicode_AsWideChar(input, buffer, 2) != 1){
                                    ok = false;
                                    return {};
                                 }
                                 return buffer[0];
                                 """,
                            to_cpp_check_precedence='SWIG_TYPECHECK_UNICHAR',
                            **type_args))

    import_(TypeSwimporting('bool', 'bool',
                            to_py="return PyBool_FromLong(input);",
                            to_cpp="""
                                if (!PyBool_Check(input)){
                                    ok = 0;
                                    return 0;
                                }
                                return (input == Py_True);
                                """,
                            to_cpp_check="return PyBool_Check(input);",
                            to_cpp_check_precedence='SWIG_TYPECHECK_BOOL',
                            **type_args))


@pools.add(TypeSwimportingPool, name='bool')
def bool_(**type_args):
    """
    Allows any python type to be interpreted as a C boolean
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    return TypeSwimporting('bool', 'Any',
                           to_cpp="""
                                   int res= PyObject_IsTrue(input);
                                   if (res == -1){
                                        ok = 0;
                                        return 0;
                                   }
                                   return res;
                                   """,
                           to_py="""
                               auto ret = input ? Py_True : Py_False;
                               Py_INCREF(ret);
                               return ret;
                               """,
                           to_cpp_check='return 1;',
                           to_cpp_check_precedence=...,
                           ref_maps=TypeSwimporting.REF_MAPS_NO_NONE_CHECK,
                           **type_args)


@pools.add(TypeSwimportingPool)
def singletons(**type_args):
    """
    the necessary typemaps for singleton use
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    yield TypeSwimporting('ellipsis', 'type(...)',
                          to_py="""
                            Py_INCREF(Py_Ellipsis);
                            return Py_Ellipsis;
                            """,
                          to_cpp="""
                            if (input != Py_Ellipsis){
                                PyErr_SetString(PyExc_TypeError,"expected ellipsis");
                                ok = false;
                                return {};
                            }
                            return Ellipsis;
                            """,
                          to_cpp_check_precedence=0,
                          **type_args
                          )
    yield TypeSwimporting('NoneType', 'type(None)',
                          to_py="""
                            Py_RETURN_NONE;
                            """,
                          to_cpp="""
                            if (input != Py_None){
                                PyErr_SetString(PyExc_TypeError,"expected None");
                                ok = false;
                                return {};
                            }
                            return None;
                            """,
                          to_cpp_check_precedence=0,
                          ref_maps=TypeSwimporting.REF_MAPS_NO_NONE_CHECK,
                          **type_args
                          )
    yield TypeSwimporting('NotImplementedType', 'type(NotImplemented)',
                          to_py="""
                            Py_RETURN_NOTIMPLEMENTED;
                            """,
                          to_cpp="""
                            if (input != Py_NotImplemented){
                                PyErr_SetString(PyExc_TypeError,"expected NotImplemented");
                                ok = false;
                                return {};
                            }
                            return NotImplemented;
                            """,
                          to_cpp_check_precedence=0,
                          **type_args
                          )
    yield TypeSwimporting('PySingleton',
                          to_py="""
                            if (input == None)
                                Py_RETURN_NONE;
                            else if (input == NotImplemented)
                                Py_RETURN_NOTIMPLEMENTED;
                            else if (input == Ellipsis)
                            {
                                Py_INCREF(Py_Ellipsis);
                                return Py_Ellipsis;
                            }
                            ok = 0;
                            Py_RETURN_NONE;
                            """,
                          to_cpp="""
                            if (input == Py_NotImplemented)
                                return NotImplemented;
                            if (input == Py_None)
                                return None;
                            if (input == Py_Ellipsis)
                                return Ellipsis;
                            PyErr_SetString(PyExc_TypeError,"expected a singleton type"); 
                            ok = false;
                            return {};
                            """,
                          to_cpp_check_precedence=0,
                          ref_maps=TypeSwimporting.REF_MAPS_NO_NONE_CHECK,
                          **type_args
                          )


@pools.add(TypeSwimportingPool, name='complex')
def complex_(**type_args):
    """
    typemaps for using std::complex<double> and std::complex<float>
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    yield TypeSwimporting('std::complex<double>', 'complex',
                          to_py="""
                                return PyComplex_FromDoubles(std::real(input), std::imag(input));
                                """,
                          to_cpp="""
                                if (PyComplex_Check(input)){
                                    return {PyComplex_RealAsDouble(input), PyComplex_ImagAsDouble(input)};
                                }
                                if (PyNumber_Check(input)){
                                    auto f = PyNumber_Float(input);
                                    auto result = PyFloat_AS_DOUBLE(f);
                                    Py_DECREF(f);
                                    return result;
                                }
                                PyErr_SetString(PyExc_TypeError, "expected a number");
                                ok = 0;
                                """,
                          to_cpp_check_precedence="SWIG_TYPECHECK_COMPLEX",
                          **type_args)
    yield TypeSwimporting('std::complex<float>', 'complex',
                          to_py="""
                                return PyComplex_FromDoubles(std::real(input), std::imag(input));
                                """,
                          to_cpp="""
                                if (PyComplex_Check(input)){
                                    return {(float)PyComplex_RealAsDouble(input), (float)PyComplex_ImagAsDouble(input)};
                                }
                                if (PyNumber_Check(input)){
                                    auto f = PyNumber_Float(input);
                                    auto result = (float)PyFloat_AS_DOUBLE(f);
                                    Py_DECREF(f);
                                    return result;
                                }
                                PyErr_SetString(PyExc_TypeError, "expected a number");
                                ok = 0;
                                """,
                          to_cpp_check_precedence="SWIG_TYPECHECK_COMPLEX",
                          **type_args)



from typing import Union

import re

from swimport.__util__ import *
from swimport.swim import Swim
from swimport.pools import syspools
from swimport.functionswim import FunctionBehaviour
from swimport.model import Variable, NameTrigger, Compileable, object_swimporting, Container, Function, SwimRule


@Variable.set_default_trigger
class VariableNameTrigger(Variable.Trigger, NameTrigger):
    """trigger for a variable object"""


class VariableTypeTrigger(Variable.Trigger):
    """
    trigger for a variable object. See methodswim.ParameterTypeTrigger for an explanation on type matching
    """

    def __init__(self, pattern: Compileable):
        self.pattern = re.compile(pattern)

    def is_valid(self, rule: 'SwimRule', obj):
        return super().is_valid(rule, obj) \
               and self.pattern.fullmatch(obj.normalized_type)


def is_const(v: Variable):
    return re.search(r'(?<!a-zA-Z_0-9)const(?!a-zA-Z_0-9)', v.normalized_type)


@Variable.set_default_behaviour
class VariableBehaviour(Variable.Behaviour):
    """
    behaviour for variable object.
    """

    def __init__(self, immutable=False, const: bool = ..., del_from_module_ns=...):
        """
        :param immutable: whether to wrap the variable in an %immutable wrap
        :param const: whether to wrap the variable in a %constant directive. default is to check for const keyword in variable type.
        :param del_from_module_ns: whether to delete the variable from the module namespace, and make it only accessible from the cvar attribute. default is to only do this for immutable variables.
        """
        self.immutable = immutable
        self.const = const
        self.del_from_module_ns = del_from_module_ns

    @object_swimporting
    def wrap(self, rule, obj: Variable, swim):
        if self.immutable:
            swim.add_raw(f'%immutable {obj.name};')

        body = obj.body
        if body.startswith('extern '):
            body = body[len('extern '):]
        const = self.const
        if const is ...:
            # disable auto-const if obj is a variable subclass (like a class member)
            const = type(obj) is Variable \
                    and is_const(obj)

        if const:
            body = '%constant ' + body

        del_ = self.del_from_module_ns
        if del_ is ...:
            del_ = self.immutable

        if del_:
            swim(syspools.ns_trap(obj.name))

        swim.add_raw(body)


class VariableGetSetBehaviour(Variable.Behaviour):
    """
    A behaviour for variable objects that wraps the variable in getter and setter functions
    """
    default_get_behaviour = FunctionBehaviour()
    default_set_behaviour = FunctionBehaviour()

    def __init__(self, get: Union[str, bool] = True, set_: Union[str, bool] = ..., name_template='{}_{!l}',
                 get_behaviour=default_get_behaviour, set_behaviour=default_set_behaviour, decl_scope=...,
                 function_scope=..., func_class = ...):
        """
        :param get: whether to create a getter method. Can also be a string to give the getter a special name.
        :param set_: whether to create a setter method, default is to only create a setter if the variable is not a
         const. Can also be a string to give the setter a special name.
        :param name_template: a formattable string to generate the names of the getter and setter methods, note that
         the template uses extended formatting, see the ExtendedFormatter class in __util__.
        :param get_behaviour: the swimport method behaviour for getter methods.
        :param set_behaviour: the swimport method behaviour for setter methods.
        :param decl_scope: the scope of the variable declaration.
        :param function_scope: the scope of the function declaration.
        """
        self.get = get
        self.set = set_
        self.name_template = name_template
        self.get_behaviour = get_behaviour
        self.set_behaviour = set_behaviour
        self.decl_scope = decl_scope
        self.function_scope = function_scope
        self.func_class = func_class

    @object_swimporting
    def wrap(self, rule: 'SwimRule', obj: Variable, swim):
        swim.add_comment(obj)
        intype = obj.type
        qual_name = obj.name
        decl_scope = self.decl_scope
        function_scope = self.function_scope
        func_class = self.func_class
        setfunction_rtn_type = 'void'

        if isinstance(obj, Container.Member) and obj.is_static:
            if intype.startswith('static '):
                intype = intype[len('static '):]
            qual_name = obj.owner.name + '::' + qual_name
            setfunction_rtn_type = 'static void'

            if decl_scope is ...:
                decl_scope = None
            if function_scope is ...:
                function_scope = Swim.add_extend

        if decl_scope is ...:
            decl_scope = Swim.add_insert
        if function_scope is ...:
            function_scope = Swim.add_inline
        if func_class is ...:
            func_class = Function.make

        if decl_scope:
            decl_scope(swim, obj.body)

        if self.get:
            if isinstance(self.get, str):
                func_name = self.get
            else:
                func_name = ExtendedFormatter().format(self.name_template, 'get', obj.name)

            get_method = func_class(func_name, [], obj.type)
            behaviour = self.get_behaviour.replace(
                body=f'{{ return {qual_name}; }}',
                body_scope=function_scope)
            swim(behaviour.do(get_method))

        mk_set = self.set

        if mk_set is ...:
            mk_set = not is_const(obj)

        if mk_set:
            if isinstance(self.set, str):
                func_name = self.set
            else:
                func_name = ExtendedFormatter().format(self.name_template, 'set', obj.name)

            set_method = Function.make(func_name, [Function.Param.make('__v', intype)], setfunction_rtn_type)
            behaviour = self.set_behaviour.replace(
                body=f'{{ {qual_name} = __v; }}',
                body_scope=function_scope)
            swim(behaviour.do(set_method))



from functools import partial

from swimport.pools.pools import pools
from swimport.pools.types import TypeSwimportingPool
from swimport.pools.derived_types.iter_ import iter_
from swimport.typeswim import TypeSwimporting


@pools.add(TypeSwimportingPool)
def array(inner_type: str, size: int, *, cpp_name='std::array<?,??>', swim, **kwargs):
    """
    as `iter`, but specialized for arrays of a specific size
    :param inner_type: the type to wrap
    :param size: the size pf the array
    :param cpp_name: the cpp type, with ?? replaced with the size
    :param kwargs: additional keyword arguments forwarded to pools.iter
    """
    cpp_name = cpp_name.replace('??', str(size))

    inner_type, inner_porting = swim.get_porting(inner_type)

    port: TypeSwimporting = partial(iter_.get,
                                    (inner_type, inner_porting),
                                    cpp_name=cpp_name, output=list)(swim=swim, insert_func='!!invalid!!', **kwargs)
    if port.to_cpp:
        body = [f"""
                ok = 1;
                auto obj_len = PyObject_Length(input);
                if (obj_len == -1){{
                    if (PyErr_Occurred() != nullptr)
                        SWIM_RAISE_EXISTING;
                }}
                else{{
                    if (obj_len != {size})
                        SWIM_RAISE_TYPE("expected {size} members in: ", input);
                }}
                auto iter = PyObject_GetIter(input);
                if (!iter){{
                    SWIM_RAISE_TYPE("expected an iterable, not ", input);
                }}
                auto temp = PyIter_Next(iter);"""]
        for i in range(size):
            body.append(f"""
            if (!temp)
            {{
                Py_DECREF(iter);
                if (PyErr_Occurred() != nullptr)
                    SWIM_RAISE_EXISTING;
                SWIM_RAISE_TYPE("expected {size} elements, got {i} in: ", input);
            }}
            {inner_type} member_{i} = {inner_porting.to_cpp_func.function_name}(temp, ok, PACKED_UD(0));
            Py_DECREF(temp);
            if (!ok || PyErr_Occurred() != nullptr){{
                Py_DECREF(iter);
                SWIM_RAISE_UNSPECIFIED;
            }}
            temp = PyIter_Next(iter);
            """)
        members = ', '.join(f'member_{i}' for i in range(size))
        body.append(f"""
        Py_DECREF(iter);
        if (temp){{
            Py_DECREF(temp);
            SWIM_RAISE_TYPE("too many elements, expected {size} in: ", input);
        }}
        else if (PyErr_Occurred() != nullptr){{
            SWIM_RAISE_EXISTING;
        }}

        return {{ {members} }};
        """)
        port.to_cpp = port.to_cpp.replace(body='\n'.join(body))
    yield port



from __future__ import annotations

from typing import Callable, Type, Any, Generic, Optional, TypeVar, List, Iterable

from abc import ABC, abstractmethod
from inspect import isabstract
from functools import update_wrapper, lru_cache
import re
from enum import Enum
from itertools import chain

import swimport.model.rules as rules_mod

import CppHeaderParser

T = TypeVar('T')

Source = Iterable['CPPObject']


class HeaderSource(Source):
    def __init__(self, header, lines):
        self._objects: List[CPPObject] = [
            *(Function(f, lines) for f in header.functions),  # methods
            *(Typedef(pair) for pair in header.typedefs.items()),  # typedefs
            *(Enumeration(e, lines) for e in header.enums),  # enums
            *(Variable(v) for v in header.variables),  # vars
            *(Container(c, lines) for c in header.classes.values())  # containers
        ]

    def __iter__(self):
        yield from self._objects


class FileSource(HeaderSource):
    """A source from a cpp header file. Reads the header file and generates CPPObject wrappers for the objects
    detected."""

    def __init__(self, source_path, directives=()):
        self.source_path = source_path
        self.directives = directives

        with open(self.source_path) as r:
            lines = [l.rstrip() for l in r]
        header = CppHeaderParser.CppHeader('\n'.join(lines), 'string')  # self.source_path, 'file')
        super().__init__(header, lines)


class RawSource(HeaderSource):
    """A source from a string"""

    def __init__(self, source_string: str):
        self.src_string = source_string

        lines = self.src_string.splitlines(False)
        header = CppHeaderParser.CppHeader(self.src_string, 'string')  # self.source_path, 'file')
        super().__init__(header, lines)


class CategorizedSurrogate(Generic[T]):
    """
    A class that holds 2 classes, one for inheriting, one for creating
    """

    def __init__(self, type_, inherit_class: Type[T]):
        self._callable: Optional[Callable[..., T]] = None
        self.owner = type_

        class InheritCls(inherit_class, specialization=type_):
            def __init_subclass__(cls, **kwargs):
                super().__init_subclass__(specialization=cls.object_category, **kwargs)

        self.inherit_cls: Type[T] = InheritCls

    def __mro_entries__(self, bases):
        return self.inherit_cls,

    def __call__(self, *args, **kwargs):
        if not self._callable:
            raise TypeError(f'this superclass surrogate is not callable'
                            f' (did you mean {self.owner.__name__}.set_default?)')
        if len(args) == 1 and not kwargs and isinstance(args[0], self):
            return args[0]
        return self._callable(*args, **kwargs)

    def __instancecheck__(self, instance):
        return isinstance(instance, self.inherit_cls)

    def __subclasscheck__(self, subclass):
        return issubclass(self, self.inherit_cls)

    @property
    def callable(self):
        return self._callable

    @callable.setter
    def callable(self, v):
        self._callable = v
        update_wrapper(self, v, assigned=('__doc__',))


class CPPObject(ABC):
    """base class for all cpp objects that can be imported"""
    Behaviour: Type['rules_mod.Behaviour']
    Trigger: Type['rules_mod.Trigger']

    @classmethod
    def set_default_behaviour(cls, func: Callable[..., 'Behaviour']):
        cls.Behaviour.callable = func
        return func

    @classmethod
    def set_default_trigger(cls, func: Callable[[Any], 'Trigger']):
        cls.Trigger.callable = func
        return func

    def __init__(self, data):
        if isinstance(data, type(self)):
            self.__init__(data._data)
        else:
            self._data = data

    @property
    @abstractmethod
    def name(self) -> str:
        """identifiable name of the object"""
        pass

    @property
    @abstractmethod
    def body(self) -> str:
        """body of the object's declaration"""
        pass

    def __repr__(self):
        return f'interface wrapper for {type(self).__name__} {self.name}'

    def __str__(self):
        return self.name

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not isabstract(cls):
            cls.Behaviour = CategorizedSurrogate(cls, rules_mod.CategorizedBehaviour)
            cls.Trigger = CategorizedSurrogate(cls, rules_mod.CategorizedTrigger)


class _Sourced(CPPObject):
    """a cpp object that also holds a reference to its own source deceleration"""

    def __init__(self, data, lines):
        super().__init__(data)
        self.source_lines = lines


class _Declaration(_Sourced):
    """a cpp object whose declarations are a single line"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._body = None

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    def body(self) -> str:
        if self._body is None:
            ind = self._data['line_number'] - 1
            ret = [self.source_lines[ind]]
            while not ret[-1].endswith(';'):
                ind += 1
                ret.append(self.source_lines[ind])
            self._body = ' '.join(r.strip() for r in ret)
        return self._body


class _Scope(_Sourced):
    """cpp object whose declarations are an entire scope"""

    @staticmethod
    def _delve(depth, maxdepth, newline):
        while newline:
            opener_index = newline.find('{')
            closer_index = newline.find('}')

            if opener_index == -1:
                if closer_index == -1:
                    break
                depth -= 1
                newline = newline[closer_index + 1:]
                continue

            if closer_index == -1 or opener_index < closer_index:
                depth += 1
                if depth > maxdepth:
                    maxdepth = depth
                newline = newline[opener_index + 1:]
                continue

            depth -= 1
            newline = newline[closer_index + 1:]
            continue
        return depth, maxdepth

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    def body(self) -> str:
        ret = []
        depth = 0
        max_depth = 0
        ind = self._data['line_number'] - 1
        while not (max_depth > 0 and depth == 0):
            ret.append(self.source_lines[ind])
            ind += 1
            depth, max_depth = self._delve(depth, max_depth, ret[-1])
        return '\n'.join(ret)


class _TypedCPPObject(CPPObject):
    """
    A cpp object that is associated with a distinct type
    """

    @property
    @abstractmethod
    def type(self) -> str:
        pass

    @property
    def normalized_type(self):
        return normalize_type(self.type)


class Function(_Declaration, _TypedCPPObject):
    """a cpp method"""

    class Param(_TypedCPPObject):
        """a function parameter, this object cannot be swimported directly"""

        def __init__(self, data, owner=None, index=None):
            super().__init__(data)
            self.owner = owner
            self.index = index

        @classmethod
        def make(cls, name, type_, default=None, **kwargs):
            data = {
                'name': name,
                'type': type_,
                'defaultValue': default
            }
            return cls(data, **kwargs)

        @property
        def name(self) -> str:
            return self._data['name']

        @property
        def type(self) -> str:
            return self._data['type']

        @property
        def default(self):
            return self._data.get('defaultValue', None)

        @property
        def body(self):
            ret = self.sign
            default = self.default
            if default:
                ret += ' = ' + default
            return ret

        @property
        def next(self) -> 'Function.Param':
            return self.owner.parameters[self.index + 1]

        @property
        def next_or_none(self) -> Optional['Function.Param']:
            try:
                return self.next
            except IndexError:
                return None

        @property
        def sign(self) -> str:
            return self.type + ' ' + self.name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params = [self.Param(p, owner=self, index=i) for (i, p) in enumerate(self._data['parameters'])]

    @classmethod
    def make(cls, name, params, rtnType):
        data = {
            'name': name,
            'rtnType': rtnType,
            'parameters': params,
        }
        ret = cls(data, None)
        ret._body = ret.sign + ';'
        return ret

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def parameters(self) -> List[Param]:
        return self.params

    @property
    def return_type(self) -> str:
        return self._data['rtnType']

    @property
    def prm_types(self):
        return (p.type for p in self.parameters)

    @property
    def prm_names(self):
        return (p.name for p in self.parameters)

    @property
    def pattern(self):
        return self.name + '(' + ', '.join(f'{p.type} {p.name}' for p in self.parameters) + ')'

    @property
    def sign(self):
        return self.return_type + ' ' + self.name + '(' + ', '.join(p.body for p in self.parameters) + ')'

    @property
    def type(self):
        return self.return_type


class Typedef(_TypedCPPObject):
    """a cpp typedef"""

    @classmethod
    def make(cls, new_type, old_type):
        return cls((new_type, old_type))

    @property
    def dest_type(self) -> str:
        return self._data[0]

    @property
    def source_type(self) -> str:
        return self._data[1]

    @property
    def body(self) -> str:
        return f'typedef {self.source_type} {self.dest_type};'

    @property
    def name(self):
        return self.dest_type

    @property
    def type(self):
        return self.source_type


class Variable(_TypedCPPObject):
    """a cpp variable"""

    const_pattern = re.compile('(?<![a-zA-Z0-9_])const(?![a-zA-Z0-9_])')

    @property
    def name(self):
        return self._data['name']

    @property
    def type(self):
        return self._data['type']

    @property
    def isextern(self):
        return self._data['extern']

    @property
    def body(self):
        preamble = []
        if self.isextern:
            preamble.append('extern')
        preamble = ' '.join(preamble)
        if preamble:
            preamble += ' '
        return preamble + self.type + ' ' + self.name + ';'


class Container(_Scope):
    """a container (class, struct, or union)"""

    def __init__(self, data, source_lines):
        if data['name'].startswith('union '):  # get rid of union header for unions
            data['name'] = data['name'][len('union '):]
        super().__init__(data, source_lines)

        self.methods = list(chain.from_iterable(
            (self.Method(m, source_lines, access=k) for m in v)
            for k, v in self._data['methods'].items()))

        self.members = list(chain.from_iterable(
            (self.Member(m, access=k, owner=self) for m in v)
            for k, v in self._data['properties'].items()))

    class Kind(Enum):
        Union = 'union'
        Class = 'class'
        Struct = 'struct'

    class Member(Variable):
        """a member of a container"""

        def __init__(self, *args, access, owner: Container):
            super().__init__(*args)
            self.access = access
            self.owner = owner

        @property
        def is_static(self):
            return bool(self._data['static'])

    class Method(Function):
        def __init__(self, *args, access):
            super().__init__(*args)
            self.access = access

        @property
        def is_static(self):
            return bool(self._data['static'])

    @property
    def name(self):
        return self._data['name']

    @property
    def kind(self) -> Kind:
        return self.Kind[self._data['declaration_method'].capitalize()]

    @property
    def public_bases(self)->Iterable[str]:
        yield from (i['class'] for i in self._data['inherits'] if i['access'] == 'public')


class Enumeration(_Declaration, _TypedCPPObject):
    """a cpp enum"""

    class EnumerationValue(CPPObject):
        """A value in a cpp enum"""

        @property
        def name(self):
            return self._data['name']

        @property
        def value(self):
            return self._data['value']

        @property
        def body(self):
            return self.name + ' = ' + str(self.value)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.members = [self.EnumerationValue(m) for m in self._data['values']]

    @property
    def name(self):
        return self._data['name']

    @property
    def type(self) -> str:
        ret = self._data['type']
        if not ret:
            return 'int'
        return getattr(ret, '__name__', ret)


idempotent_qualifiers = frozenset(('const', 'volatile', 'mutable'))
prefix_qualifiers = ('const', 'volatile', 'mutable')
prefix_pattern = re.compile('^(' + '|'.join(
    (re.escape(q) + (r'(?![a-zA-Z0-9_])' if q[-1].isalnum() else ''))
    for q in prefix_qualifiers) + ')')
postfix_qualifiers = ('const', 'volatile', 'mutable', '*', '&&', '&')
postfix_pattern = re.compile('(' + '|'.join(
    ((r'(?<![a-zA-Z0-9_])' if q[0].isalnum() else '') + re.escape(q))
    for q in postfix_qualifiers) + ')$')


@lru_cache()
def normalize_type(type_str: str):
    """
    Normalizes a c++ type string in the following ways:
    * qualifiers are gathered and are always ordered right-to-left, space separated
        (const int *  -> int const *)
    * primitive type name are normalized to their simplest versions:
        (signed short int -> short)
    * const, volatile, and mutable qualifiers are bunched, and sorted alphabetically
        (volatile const mutable const volatile int -> int const mutable volatile)

    >>> normalize_type('int')
    'int'
    >>> normalize_type('const int')
    'int const'
    >>> normalize_type('int const *')
    'int const *'
    >>> normalize_type('   char const*    ')
    'char const *'
    >>> normalize_type('const wchar_t &const*')
    'wchar_t const & const *'
    >>> normalize_type('signed   long            long   int')
    'long long'
    >>> normalize_type('signed short    int')
    'short'
    >>> normalize_type('unsigned short    int')
    'unsigned short'
    >>> normalize_type('int&& const')
    'int && const'
    >>> normalize_type('const short int*&')
    'short const * &'
    >>> normalize_type('long      double')
    'long double'
    >>> normalize_type('const char * const volatile')
    'char const * const volatile'
    >>> normalize_type('char * volatile const')
    'char * const volatile'
    >>> normalize_type('char * const volatile const')
    'char * const volatile'
    >>> normalize_type('const char const const const *')
    'char const *'
    >>> normalize_type('char**')
    'char * *'
    >>> normalize_type('const constintvolatile*&')
    'constintvolatile const * &'
    >>> normalize_type('volatile const mutable const volatile int')
    'int const mutable volatile'
    >>> normalize_type('int*const&volatile*const&&const**')
    'int * const & volatile * const && const * *'
    """
    stack = []
    type_str = type_str.strip()
    # gather the components
    while True:
        match = postfix_pattern.search(type_str) or prefix_pattern.match(type_str)
        if not match:
            break
        kw = match.group(1)
        stack.append(kw)
        if match.re is postfix_pattern:
            type_str = type_str[:-len(kw)].rstrip()
        else:  # match.re is cls.prefix_pattern
            type_str = type_str[len(kw):].lstrip()

    # at this point, type_str holds the primitive name (int, long int, unsigned short,...)
    # we need to normalize it
    type_str = re.sub(r'\s{2,}', ' ', type_str.strip())
    if type_str in ('signed', 'unsigned'):
        type_str += ' int'

    if type_str.startswith('signed ') \
            and any(type_str.endswith(' ' + pi) for pi in ('int', 'short', 'long')):
        type_str = type_str[len('signed '):]

    for actual_type in ('short', 'long', 'long long'):
        if type_str.endswith(' ' + actual_type + ' int') \
                or type_str == actual_type + ' int':
            type_str = type_str[:-len(' int')]
            break

    # now type_str holds the normalized primitive type, add qualifiers to it
    ret = [type_str]

    # all the idempotent, order invariant qualifiers for the current symbol (const, volatile...)
    idempotent_qualifier_subset = set()

    for qualifier in reversed(stack):
        if qualifier in idempotent_qualifier_subset:
            continue  # const const -> const
        if qualifier in idempotent_qualifiers:
            idempotent_qualifier_subset.add(qualifier)
        else:
            for q in sorted(idempotent_qualifier_subset):
                ret.append(q)
            idempotent_qualifier_subset.clear()
            ret.append(qualifier)
    for q in sorted(idempotent_qualifier_subset):
        ret.append(q)
    return ' '.join(ret)
